import random
from django.db import models
from simulations.utils import throw_dice

from django.conf import settings


class Experiment(models.Model):
    class NetworkTopology(models.TextChoices):
        SCALE_FREE = "Scale Free"
        WEIGHTED_SCALE_FREE = "Weighted Scale Free"
        SMALL_WORLD = "Small World"
        WEIGHTED_SMALL_WORLD = "Weighted Small World"
        CLUSTERED_SMALL_WORLD = "Clustered Small World"
        WEIGHTED_CLUSTERED_SMALL_WORLD = "Weighted Clustered Small World"

    class MitigationFactorFunction(models.TextChoices):
        STEP_WISE = "Step Wise"
        LINEAR = "Linear"

    class PopulationHeterogeneity(models.TextChoices):
        HOMOGENEOUS = "Homogeneous"
        HETEROGENEOUS = "Heterogeneous"

    started_at = models.DateTimeField(auto_now_add=True)
    finished_at = models.DateTimeField(default=None, blank=True, null=True)
    label = models.CharField(max_length=100, blank=True, null=True)
    network_topology = models.CharField(
        max_length=100, null=False, blank=False, choices=NetworkTopology.choices
    )

    population_size = models.IntegerField(default=1000)
    number_of_generations = models.IntegerField(default=10)
    number_of_time_steps = models.IntegerField(default=1000)
    generation_size = models.IntegerField(default=50)
    immune_system_base_string = models.CharField(max_length=10)
    immunity_duration = models.IntegerField(default=5, null=True, blank=True)

    immune_response_enabled = models.BooleanField(default=True)
    weighted_connections_enabled = models.BooleanField(default=True)
    isolation_if_sick_enabled = models.BooleanField(default=False)
    static_immune_system_enabled = models.BooleanField(
        default=True,
        help_text="If enabled, all individuals will share the same string for their immune system",
    )
    patient_zero_strategy = models.CharField(
        max_length=100,
        null=True,
        blank=True,
    )
    mitigation_factor_function = models.CharField(
        max_length=100,
        null=True,
        blank=True,
        choices=MitigationFactorFunction.choices,
    )
    population_heterogeneity = models.CharField(
        max_length=100,
        null=True,
        blank=True,
        choices=PopulationHeterogeneity.choices,
    )

    performance_function_death_weight = models.FloatField(
        default=4.0, null=False, blank=False
    )
    performance_function_infection_weight = models.FloatField(
        default=1.0, null=False, blank=False
    )
    performance_function_recovery_weight = models.FloatField(
        default=-1.0, null=False, blank=False
    )

    base_individual_network_topology = models.ManyToManyField(
        "simulations.Individual",
        related_name="base_individual_network_topology",
    )

    # Statistics
    number_of_nodes = models.IntegerField(default=0)
    number_of_edges = models.IntegerField(default=0)
    average_degree = models.FloatField(default=0.0)
    average_clustering_coefficient = models.FloatField(default=0.0)
    average_shortest_path_length = models.FloatField(default=0.0)
    diameter = models.IntegerField(default=0)
    average_degree_centrality = models.FloatField(default=0.0)
    average_betweenness_centrality = models.FloatField(default=0.0)
    average_closeness_centrality = models.FloatField(default=0.0)
    average_eigenvector_centrality = models.FloatField(default=0.0)
    node_connectivity = models.IntegerField(default=0)
    edge_connectivity = models.IntegerField(default=0)

    @property
    def runtime(self):
        if not self.finished_at:
            return None
        return self.finished_at - self.started_at

    @property
    def runtime_hours(self):
        if not self.runtime:
            return "N/A"
        runtime_hours = self.runtime.seconds / 3600
        decimal_hours = runtime_hours % 1
        minutes = decimal_hours * 60
        return f"{runtime_hours:.0f} hour {minutes:.0f}min"

    @property
    def get_best_perfomer(self):
        return self.generation_set.all().order_by("-pathogen_performance_score").first()


class Generation(models.Model):
    experiment = models.ForeignKey(
        "simulations.Experiment", on_delete=models.CASCADE, null=True
    )
    generation_number = models.IntegerField(default=1)
    base_pathogen = models.OneToOneField(
        "simulations.Pathogen",
        on_delete=models.CASCADE,
        null=False,
        related_name="generation",
        help_text="Base pathogen, used to infect patient zero",
    )

    # Statistics
    number_of_infections = models.IntegerField(default=0)
    number_of_deaths = models.IntegerField(default=0)
    number_of_recoveries = models.IntegerField(default=0)
    pathogen_performance_score = models.FloatField(default=0.0)

    @property
    def base_pathogen_mitigation_factor(self):
        return calculate_mitigation_factor(
            self.experiment.immune_system_base_string, self.base_pathogen.attack_vector
        )


class Connection(models.Model):
    # Should rename the related names here. From too something
    individual = models.ForeignKey(
        "simulations.Individual",
        on_delete=models.CASCADE,
        related_name="individual_connections",
    )
    connected_individual = models.ForeignKey(
        "simulations.Individual",
        on_delete=models.CASCADE,
        related_name="connected_individuals",
    )
    weight = models.FloatField(default=1.0)


class Individual(models.Model):
    class StateOptions(models.TextChoices):
        ALIVE = "Alive"
        DEAD = "Dead"

    IMMUNE_SYSTEM_STRING_LENGTH = 10

    connections = models.ManyToManyField(
        "simulations.Individual", blank=True, through="simulations.Connection"
    )

    generation = models.ForeignKey(
        "simulations.Generation",
        on_delete=models.CASCADE,
        null=True,
        related_name="individuals",
    )
    # Subject to change on genotype length
    pathogen = models.OneToOneField(
        "simulations.Pathogen",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        related_name="individual",
    )
    state = models.CharField(
        max_length=16,
        null=False,
        blank=False,
        choices=StateOptions.choices,
        default=StateOptions.ALIVE,
    )
    immune_system = models.CharField(
        max_length=IMMUNE_SYSTEM_STRING_LENGTH, null=False, blank=False
    )

    @classmethod
    def init_individuals(cls, population, population_size=1000):
        individuals = cls.objects.bulk_create(
            [
                Individual(
                    population=population,
                    immune_system="".join(
                        [str(random.randint(0, 1)) for _ in range(10)]
                    ),
                )
                for _ in range(population_size)
            ]
        )
        return individuals


# Is there a benefit to encoding the attack vector as a DNA sequence? Probably not
def bitstring_to_dna_sequence(bitstring: str) -> str:
    """
    Convert a bitstring to a DNA sequence.

    :param bitstring: A string of 2-bit groups
    :return: A string of nucleotides
    """
    # Define a mapping of 2-bit groups to nucleotides
    bit_to_nucleotide = {
        "00": "A",
        "01": "C",
        "10": "G",
        "11": "T",
    }

    # Convert the bitstring to a DNA sequence
    dna_sequence = "".join(
        [bit_to_nucleotide[bitstring[i : i + 2]] for i in range(0, len(bitstring), 2)]
    )

    return dna_sequence


class Pathogen(models.Model):
    # Skill points that are translated into rates depending on min-max values
    infection_rate_points = models.IntegerField()
    death_rate_points = models.IntegerField()
    mutation_rate_points = models.IntegerField()
    incubation_time_points = models.IntegerField()
    infection_time_points = models.IntegerField()

    MAXIMUM_RATE = 1.0
    MINIMUM_RATE = 0.0

    # Need to play around with these values a bit more
    MINIMUM_TIME = 0
    MAXIMUM_TIME = 10

    days_infected = models.IntegerField(default=0)
    attack_vector = models.CharField(max_length=10)

    @property
    def skill_points_spent(self):
        return (
            self.infection_rate_points
            + self.death_rate_points
            + self.mutation_rate_points
            + self.incubation_time_points
            + self.infection_time_points
        )

    # Properties calculating from skill points
    @property
    def infection_rate(self):
        # Normalize the infection_rate_points to the range [MINIMUM_RATE, MAXIMUM_RATE]
        normalized_points = self.infection_rate_points / self.skill_points_spent
        return self.MAXIMUM_RATE * normalized_points + self.MINIMUM_RATE * (
            1 - normalized_points
        )

    @property
    def death_rate(self):
        # Normalize the death_rate_points to the range [MINIMUM_RATE, MAXIMUM_RATE]
        normalized_points = self.death_rate_points / self.skill_points_spent
        return self.MAXIMUM_RATE * normalized_points + self.MINIMUM_RATE * (
            1 - normalized_points
        )

    @property
    def mutation_rate(self):
        # Normalize the mutation_rate_points to the range [MINIMUM_RATE, MAXIMUM_RATE]
        normalized_points = self.mutation_rate_points / self.skill_points_spent
        return self.MAXIMUM_RATE * normalized_points + self.MINIMUM_RATE * (
            1 - normalized_points
        )

    @property
    def incubation_time(self):
        # Normalize the incubation_time_points to the range [MINIMUM_TIME, MAXIMUM_TIME]
        normalized_points = self.incubation_time_points / self.skill_points_spent
        return int(
            self.MAXIMUM_TIME * normalized_points
            + self.MINIMUM_TIME * (1 - normalized_points)
        )

    @property
    def infection_time(self):
        # Normalize the infection_time_points to the range [MINIMUM_TIME, MAXIMUM_TIME]
        normalized_points = self.infection_time_points / self.skill_points_spent
        return int(
            self.MAXIMUM_TIME * normalized_points
            + self.MINIMUM_TIME * (1 - normalized_points)
        )

    def _int_to_bitstring(self, value, length):
        return format(value, f"0{length}b")

    def infect(self, individual):
        # Copy over the attributes into a new instance of the pathogen
        new_pathogen = self.__class__.objects.create(
            infection_rate_points=self.infection_rate_points,
            death_rate_points=self.death_rate_points,
            mutation_rate_points=self.mutation_rate_points,
            incubation_time_points=self.incubation_time_points,
            infection_time_points=self.infection_time_points,
            attack_vector=self.attack_vector,
        )

        # Need to decide how this function should work
        if throw_dice(self.mutation_rate) and False:

            new_pathogen.mutate(1)
            # Mutate attack vector ?

        individual.pathogen = new_pathogen
        self.save()
        generation = individual.generation
        generation.number_of_infections += 1
        generation.save()
        individual.save()

    @property
    def is_active(self):
        return self.days_infected < self.infection_time + self.incubation_time

    @property
    def is_sick(self):

        return (
            self.incubation_time
            < self.days_infected
            < self.incubation_time + self.infection_time
        )

    def _terminate_check(self):
        immunity_duration = self.individual.generation.experiment.immunity_duration

        if immunity_duration is None:
            return

        total_days_active = self.infection_time + self.incubation_time
        immunity_consumed = self.days_infected - total_days_active

        if immunity_consumed > immunity_duration:
            # Need to properly delete so we don't have a dangling pathogen
            generation = self.individual.generation
            generation.number_of_recoveries += 1
            generation.save()
            Individual.objects.filter(pathogen=self).update(pathogen=None)
            Pathogen.objects.filter(id=self.id).delete()

    def tick(self):
        """
        The main infection loop. Calling this is equivalent with moving
        the infection forward by one day.
        """

        try:
            self.individual
        except Individual.DoesNotExist:
            raise ValueError("Pathogen is not attached to an individual")
        except Exception:
            raise

        if self.individual.state == Individual.StateOptions.DEAD:
            return

        self.days_infected += 1
        self.save()

        if self.is_sick:
            if self.individual.generation.experiment.immune_response_enabled:
                mitigation_factor = calculate_mitigation_factor(
                    self.individual.immune_system, self.attack_vector
                )
            else:
                mitigation_factor = 1

            if throw_dice(self.death_rate * mitigation_factor):
                self.individual.state = Individual.StateOptions.DEAD
                self.individual.save()
                self.individual.generation.number_of_deaths += 1
                self.individual.generation.save()
                return

        self._terminate_check()

    def get_attack_vector_dna_sequence(self) -> str:
        return bitstring_to_dna_sequence(self.attack_vector)

    @property
    def attribute_dna_sequence(self) -> str:
        """ """
        point_attribute_list = [
            "infection_rate_points",
            "death_rate_points",
            "mutation_rate_points",
            "incubation_time_points",
            "infection_time_points",
        ]
        binary_string = ""
        for attribute in point_attribute_list:
            binary_string += self._int_to_bitstring(
                getattr(self, attribute), 8
            )  # 100 in binary has length 7 (maxvalue). But nucleotides force us to make it divisible by 4

        return bitstring_to_dna_sequence(binary_string)


# TODO: Reduce the overlap threshold
def calculate_mitigation_factor(
    immune_system, attack_vector, strategy=Experiment.MitigationFactorFunction.STEP_WISE
):
    """
    Calculates the mitigation factor of the immune system against the
    attack vector.

    Assumes equal length of the two strings. Returns a value between
    0 and 1, where 0 means no mitigation and 1 means full immunity.
    """

    if type(immune_system) != str or type(attack_vector) != str:
        raise TypeError("Both arguments must be strings")

    if len(immune_system) != len(attack_vector):
        raise ValueError("The two strings must be of equal length")

    immune_system_int = int(immune_system, 2)
    attack_vector_int = int(attack_vector, 2)

    # Perform bitwise AND operation
    overlap = bin(immune_system_int & attack_vector_int).count("1")

    # Calculate mitigation factor
    mitigation_ratio = overlap / len(immune_system)

    if strategy == Experiment.MitigationFactorFunction.LINEAR:
        return mitigation_ratio
    elif strategy == Experiment.MitigationFactorFunction.STEP_WISE:

        # This is too strict. Should probably add another level to this.
        # Also need to figure out how the mutation mechanism is supposed to work.
        if mitigation_ratio == 1.0:
            return 0
        elif mitigation_ratio >= 0.9:
            return 0.1
        elif mitigation_ratio >= 0.7:
            return 0.25
        elif mitigation_ratio >= 0.5:
            return 0.5
        elif mitigation_ratio >= 0.3:
            return 0.75
        else:
            return 1
    else:
        raise ValueError("Invalid mitigation factor strategy")
