import random
from enum import Enum

from django.db.models import Sum, Count


def generate_scale_free_population(population_or_individuals, weighted=False):
    from simulations.models import Individual

    """
    Accepts an uncoupled population and generates a scale free network topology.

    A scale free network is a network where the degree distribution follows a power law.
    This means that the number of connections an individual has is proportional to the
    number of connections they have. This is a good representation of how real world
    networks are structured.
    """
    individuals = population_or_individuals

    # We first pop the first two individuals and connect them
    first_individual = individuals[0]
    second_individual = individuals[1]

    # Forward and backward relation is not merged. Look into this
    first_individual.connections.add(second_individual)
    second_individual.connections.add(first_individual)
    total_connections = 1
    connection_pool = {
        f"{first_individual.id}": 1,
        f"{second_individual.id}": 1,
    }

    # individuals = population.individuals.all()

    for index, individual in enumerate(individuals[2:]):
        random_individual = random.choices(
            list(connection_pool.keys()),
            k=1,
            weights=[i / total_connections for i in connection_pool.values()],
        )[0]

        random_individual_object = Individual.objects.get(id=random_individual)

        weight = 1
        if weighted:
            weight = random.random()

        # Connect the individual to the random individual
        individual.connections.add(
            random_individual_object, through_defaults={"weight": weight}
        )
        random_individual_object.connections.add(individual)

        # Update the connection pool
        connection_pool[random_individual] += 1
        connection_pool[f"{individual.id}"] = 1

        total_connections += 1

        # connection_pool.(random_individual)
        # connection_pool_connection_count[]

        # print(f"Chose individual {random_individual.id} with {random_individual.num_connections} connections")
        # print(f"The probability was {random_individual.num_connections / total_connections}")

        # Connect the individual to the random individual
        # individual.connections.add(random_individual)
        # random_individual.connections.add(individual)

    # for value in (
    #    individuals.all()
    #    .annotate(num_connections=Count("connections"))
    ##    .order_by("-num_connections")
    #    .values_list("num_connections", "id")
    # ):
    #    print(value)

    # print(total_connections)
    # print(individuals.count())


def generate_small_world_population(
    population_or_individuals, k=4, p=0.01, weighted=False
):
    """
    Accepts an uncoupled population and generates a small world network topology.
    """

    individuals = population_or_individuals

    # Create a ring lattice
    for i in range(len(individuals)):
        for j in range(1, k // 2 + 1):
            target = (i + j) % len(individuals)
            weight = 1
            if weighted:
                weight = random.random()
            individuals[i].connections.add(
                individuals[target], through_defaults={"weight": weight}
            )

    # Rewire edges with probability p
    for i in range(len(individuals)):
        for j in range(1, k // 2 + 1):
            target = (i + j) % len(individuals)
            if random.random() < p:
                # Remove the existing connection
                individuals[i].connections.remove(individuals[target])

                # Choose a random new connection
                new_target = random.choice(individuals)
                while (
                    new_target == individuals[i]
                    or new_target in individuals[i].connections.all()
                ):
                    new_target = random.choice(individuals)

                # Add the new connection

                weight = 1
                if weighted:
                    weight = random.random()
                individuals[i].connections.add(
                    new_target, through_defaults={"weight": weight}
                )


def generate_clustered_small_world_population(
    n_clusters,
    population_or_individuals,
    cluster_size=50,
    k=4,
    p=0.05,
    inter_cluster_p=0.05,
    weighted=False,
):
    """
    Accepts an uncoupled population and generates a clustered small world network topology.

    TODO: ON OCCASION THIS CREATES AN UNCONNECTED NETWORK. FIX THIS
    """

    individuals = population_or_individuals

    # Check if there are enough individuals for the specified clusters and cluster size
    assert (
        len(individuals) >= n_clusters * cluster_size
    ), "Not enough individuals for the specified number of clusters and cluster size"

    # Create the clusters
    for cluster in range(n_clusters):
        start = cluster * cluster_size
        end = start + cluster_size
        cluster_individuals = individuals[start:end]

        # Create a ring lattice for each cluster
        for i in range(cluster_size):
            for j in range(1, k // 2 + 1):
                target = (i + j) % cluster_size
                weight = 1
                if weighted:
                    weight = random.random()
                cluster_individuals[i].connections.add(
                    cluster_individuals[target], through_defaults={"weight": weight}
                )

        # Rewire edges with probability p within each cluster
        for i in range(cluster_size):
            for j in range(1, k // 2 + 1):
                target = (i + j) % cluster_size
                if random.random() < p:
                    # Remove the existing connection
                    cluster_individuals[i].connections.remove(
                        cluster_individuals[target]
                    )

                    # Create a list of potential targets
                    potential_targets = list(cluster_individuals)

                    # Remove the current individual from the potential targets
                    potential_targets.remove(cluster_individuals[i])

                    # Remove any individuals that are already connected
                    potential_targets = [
                        individual
                        for individual in potential_targets
                        if individual not in cluster_individuals[i].connections.all()
                    ]

                    # If there are no valid targets, continue to the next individual
                    if not potential_targets:
                        continue

                    # Choose a random new connection
                    new_target = random.choice(potential_targets)

                    # Add the new connection
                    weight = 1
                    if weighted:
                        weight = random.random()
                    cluster_individuals[i].connections.add(
                        new_target, through_defaults={"weight": weight}
                    )

    # Connect clusters together with probability inter_cluster_p
    for i in range(len(individuals)):
        if random.random() < inter_cluster_p:
            # Create a list of potential targets
            potential_targets = list(individuals)

            # Remove the current individual from the potential targets
            potential_targets.remove(individuals[i])

            # Remove any individuals that are already connected
            potential_targets = [
                individual
                for individual in potential_targets
                if individual not in individuals[i].connections.all()
            ]

            # If there are no valid targets, continue to the next individual
            if not potential_targets:
                continue

            # Choose a random new connection
            new_target = random.choice(potential_targets)

            # Add the new connection
            weight = 1
            if weighted:
                weight = random.random()
            individuals[i].connections.add(
                new_target, through_defaults={"weight": weight}
            )


class ConnectionWeightClasses(Enum):
    """
    Chapter 2 - Human Social Networks - Valerio Arnaboldi, Andrea Passarella, Marco Conti, Robin I.M. Dunbar¨

    In a modern society this should probably be inverted in some way. Physical interactions often rely
    on people who see each other daily, typically work-relations which naturally would be a part
    of the active network. This model makes a bit more sense if we consider the emotional component
    and how strong those connections are.
    """

    SUPPORT_CLIQUE = 0.8
    SYMPATHY_GROUP = 0.5
    AFFINITY_GROUP = 0.3
    ACTIVE_NETWORK = 0.1


def generate_ego_network_population(population, max_layers=3, base_connections=3):
    """
    Accepts an uncoupled population and generates an ego network topology.

    We allow for the following limits on the connections
    Support clique: 3-5
    Sympathy group: 9-15
    Affinity group: 27-45
    Active network: 81-135

    Upper limit should be around 150.

    We can recursively expand each individual's network. When the total number of connections within
    a layer is exhausted we go to the next, if we have reached the max number of layers we stop and
    then find an incomplete ego in the whole network and then expand that one.
    """
    raise NotImplementedError("This is not implemented yet")

    individuals = population.individuals.all()
    connection_counts = {individual.id: 0 for individual in individuals}
    connection_weight_classes = list(ConnectionWeightClasses)

    def select_n_unique_nodes(individuals, num_nodes, exclude):
        nodes = [i for i in individuals if i not in exclude]
        return random.sample(nodes, min(num_nodes, len(nodes)))

    def connect_nodes_with_weight(current_node, new_node, weight):
        current_node.connections.add(new_node, through_defaults={"weight": weight})
        new_node.connections.add(current_node, through_defaults={"weight": weight})
        connection_counts[current_node.id] += 1
        connection_counts[new_node.id] += 1

    def expand_ego_network(ego, max_layers, base_connections):
        current_layer_nodes = [ego]
        for layer, weight_class in zip(
            range(1, max_layers + 1), connection_weight_classes
        ):
            next_layer_nodes = []

            for current_node in current_layer_nodes:
                num_connections = min(
                    base_connections * (3 ** (layer - 1)),
                    150 - connection_counts[current_node.id],
                )
                new_nodes = select_n_unique_nodes(
                    individuals, num_connections, exclude=[ego] + next_layer_nodes
                )

                for new_node in new_nodes:
                    if connection_counts[new_node.id] < 150:
                        weight = weight_class.value
                        connect_nodes_with_weight(current_node, new_node, weight)

                next_layer_nodes.extend(new_nodes)

            current_layer_nodes = next_layer_nodes

    # Expand ego networks until all nodes have reached the maximum number of connections (150)
    while any(count < 150 for count in connection_counts.values()):
        incomplete_egos = [i for i in individuals if connection_counts[i.id] < 150]
        if not incomplete_egos:
            break
        ego = random.choice(incomplete_egos)
        expand_ego_network(ego, max_layers, base_connections)


class PatientZeroStrategies(Enum):
    RANDOM = "RANDOM"
    MOST_CONNECTIONS = "MOST_CONNECTIONS"
    LEAST_CONNECTIONS = "LEAST_CONNECTIONS"


def get_patient_zero(generation, strategy=PatientZeroStrategies.MOST_CONNECTIONS):
    """
    Returns the patient zero of a population based on the strategy.
    """
    if strategy == PatientZeroStrategies.RANDOM:
        return generation.individuals.order_by("?").first()

    elif strategy == PatientZeroStrategies.MOST_CONNECTIONS:
        return (
            generation.individuals.all()
            .annotate(num_connections=Count("connections"))
            .order_by("-num_connections")
            .first()
        )

    elif strategy == PatientZeroStrategies.LEAST_CONNECTIONS:
        # Probably never going to use this. Will in most cases lead to early termination of the simulation.
        return (
            generation.individuals.all()
            .annotate(num_connections=Count("connections"))
            .order_by("num_connections")
            .first()
        )
    # Todo add strategy for fewest connection and farthest away from others
    else:
        raise ValueError("Unknown strategy")
