from simulations.population_statistics import (
    generation_statistics,
)
from simulations.models import (
    Pathogen,
    Individual,
    calculate_mitigation_factor,
)
from simulations.utils import throw_dice

from django.db.models import F, ExpressionWrapper, IntegerField
from django.conf import settings


def generation_tick(generation):
    # Compute incubation_time and infection_time fields using the point system

    # This assumes 100 skill points
    computed_incubation_time = ExpressionWrapper(
        F("incubation_time_points") / 100.0 * Pathogen.MAXIMUM_TIME
        + Pathogen.MINIMUM_TIME * (1 - F("incubation_time_points") / 100.0),
        output_field=IntegerField(),
    )

    computed_infection_time = ExpressionWrapper(
        F("infection_time_points") / 100.0 * Pathogen.MAXIMUM_TIME
        + Pathogen.MINIMUM_TIME * (1 - F("infection_time_points") / 100.0),
        output_field=IntegerField(),
    )

    # Retrieve pathogens of living individuals
    all_active_pathogens = (
        Pathogen.objects.filter(
            individual__generation=generation,
            individual__state=Individual.StateOptions.ALIVE,
        )
        .annotate(
            computed_incubation_time=computed_incubation_time,
            computed_infection_time=computed_infection_time,
        )
        .filter(
            days_infected__lt=F("computed_incubation_time")
            + F("computed_infection_time")
        )
    )

    # Get pathogens with immunity and increment days infected by one. They are
    # excluded in the query above. But what about the termination check?
    # Does not seem to trigger anymore
    Pathogen.objects.filter(
        individual__generation=generation,
        individual__state=Individual.StateOptions.ALIVE,
        individual__isnull=False,
    ).annotate(
        computed_incubation_time=computed_incubation_time,
        computed_infection_time=computed_infection_time,
    ).filter(
        days_infected__gte=F("computed_incubation_time") + F("computed_infection_time")
    ).update(
        days_infected=F("days_infected") + 1
    )

    terminate_check = (
        Pathogen.objects.filter(
            individual__generation=generation,
            individual__state=Individual.StateOptions.ALIVE,
            individual__isnull=False,
        )
        .annotate(
            computed_incubation_time=computed_incubation_time,
            computed_infection_time=computed_infection_time,
        )
        .filter(
            days_infected__gte=F("computed_incubation_time")
            + F("computed_infection_time")
        )
    )

    for path in terminate_check:
        # Kinda wonky but works for now
        path._terminate_check()

    print(f"Retrieving pathogens: {all_active_pathogens.count()}")

    for pathogen in all_active_pathogens:
        pathogen.tick()

        if not pathogen.is_active:
            continue

        infected = pathogen.individual

        if generation.experiment.isolation_if_sick_enabled and pathogen.is_sick:
            # Individual is experiencing symptoms and isolate themselves
            continue

        infected_connection = (
            infected.connected_individuals.all() | infected.individual_connections.all()
        )
        for connection in infected_connection.distinct():
            if connection.individual == infected:
                # not the recipient
                recipient = connection.connected_individual
            else:
                recipient = connection.individual

            if recipient.pathogen:
                # Already infected
                continue
            # We use weight to determine if they have had an encounter
            # A weight is a value < w <= 1. 1 Implies they have a 100% chance of having an encounter

            # We set the inequality to be smaller than the weight so the mental model works
            if generation.experiment.weighted_connections_enabled:
                dice_chance = connection.weight
            else:
                dice_chance = 1.0

            if throw_dice(dice_chance):
                # We have an encounter, now we need to determine if the pathogen is transmitted

                # We use the infection rate to determine if the pathogen is transmitted
                if generation.experiment.immune_response_enabled:
                    mitigating_factor = calculate_mitigation_factor(
                        infected.immune_system, infected.pathogen.attack_vector
                    )
                else:
                    mitigating_factor = 1

                if throw_dice(pathogen.infection_rate * mitigating_factor):
                    pathogen.infect(recipient)

    return generation_statistics(generation)
