import random

from django.core.exceptions import ValidationError
from django.db.models import Q, Count
from django.test import TestCase
from simulations.models import (
    Simulation,
    Population,
    Individual,
    Pathogen,
    Connection,
    Experiment,
    Generation,
)
from simulations.network_strategies import (
    generate_scale_free_population,
    generate_small_world_population,
    generate_ego_network_population,
    PatientZeroStrategies,
    get_patient_zero,
)
from simulations.model_serializers import (
    deserialize_individuals_and_connections,
    serialize_individuals_and_connections,
)
from simulations.visualization import create_text_bar_chart


class NetworkStrategyTests(TestCase):
    def setUp(self) -> None:
        self.simulation = Simulation.objects.create()
        self.population = Population.objects.create(
            network_topology=Population.NetworkTopology.SCALE_FREE,
            simulation=self.simulation,
        )
        self.experiment = Experiment.objects.create()
        self.generation = Generation.objects.create(experiment=self.experiment)

        self.individuals = Individual.objects.bulk_create(
            [
                Individual(
                    population=None,
                    generation=self.generation,
                    immune_system="".join(
                        [
                            str(random.randint(0, 1))
                            for _ in range(Individual.IMMUNE_SYSTEM_STRING_LENGTH)
                        ]
                    ),
                )
                for _ in range(1000)
            ]
        )

    def test_scale_free(self):
        generate_scale_free_population(self.population)

        self.assertEqual(self.population.individuals.count(), 1000)

    def test_small_world(self):
        generate_small_world_population(self.population)

        self.assertEqual(self.population.individuals.count(), 1000)

    def test_serialize_networks(self):
        generate_scale_free_population(self.individuals)
        individuals = self.generation.individuals.all()
        out = serialize_individuals_and_connections(individuals)
        new_generation = Generation.objects.create()
        deserialize_individuals_and_connections(*out, new_generation)

        connections = Connection.objects.filter(
            Q(individual__generation=new_generation)
            | Q(connected_individual__generation=new_generation)
        ).distinct()

        old_connections = Connection.objects.filter(
            Q(individual__generation=self.generation)
            | Q(connected_individual__generation=self.generation)
        ).distinct()

        self.assertEqual(
            self.generation.individuals.count(), new_generation.individuals.count()
        )
        self.assertEqual(connections.count(), old_connections.count())

        # Ordering by number of connections should be the same
        old_population_individuals = self.generation.individuals.annotate(
            num_connections=Count("individual_connections")
        ).order_by("-num_connections")

        new_population_individuals = new_generation.individuals.annotate(
            num_connections=Count("individual_connections")
        ).order_by("-num_connections")

        for old_individual, new_individual in zip(
            old_population_individuals, new_population_individuals
        ):
            self.assertEqual(
                old_individual.num_connections, new_individual.num_connections
            )

    def test_ego_network(self):
        pass


class PathogenTests(TestCase):
    def setUp(self) -> None:
        self.simulation = Simulation.objects.create()
        self.population = Population.objects.create(
            network_topology=Population.NetworkTopology.SMALL_WORLD,
            simulation=self.simulation,
        )
        self.individuals = Individual.init_individuals(self.population, 100)
        generate_scale_free_population(self.population)

        self.pathogen = Pathogen.objects.create(
            infection_rate=0.8,
            incubation_time=5,
            infection_time=10,
            death_rate=0.05,
            mutation_rate=0.5,
            individual=self.population.individuals.first(),
            days_infected=0,
        )

    def test__pathogen_tick__increases_infected_time(self):
        days_infected = self.pathogen.days_infected
        self.pathogen.tick()
        day_diff = self.pathogen.days_infected - days_infected
        self.assertEqual(day_diff, 1)

    def test__pathogen_tick__self_terminates_correctly(self):
        self.pathogen.days_infected = 10
        self.pathogen.incubation_time = 5
        self.pathogen.infection_time = 5
        self.pathogen.immunity_duration = 0
        pathogen_id = self.pathogen.id

        self.pathogen.tick()
        self.assertFalse(Pathogen.objects.filter(id=pathogen_id).exists())

    def test__pathogen_tick_no_immunity_duration__never_terminates(self):
        self.pathogen.days_infected = 1000
        self.incubation_time = 5
        self.infection_time = 5
        self.immunity_duration = 0
        pathogen_id = self.pathogen.id

        self.pathogen.tick()
        self.assertTrue(Pathogen.objects.filter(id=pathogen_id).exists())


class TesttSimulation(TestCase):
    def setUp(self) -> None:
        population_size = 10000
        self.days = 60

        self.simulation = Simulation.objects.create()
        self.population = Population.objects.create(
            network_topology=Population.NetworkTopology.SCALE_FREE,
            simulation=self.simulation,
        )

        self.individuals = Individual.init_individuals(self.population, population_size)
        generate_scale_free_population(self.population)
        self.pathogen = Pathogen.objects.create(
            infection_rate=0.5,
            incubation_time=3,
            infection_time=6,
            death_rate=0.05,
            mutation_rate=0.0,
            days_infected=0,
            immunity_duration=10,
        )
        # Get patient zero from person with largest count of connections
        # patient_zero = self.population.individuals.order_by("-connections__count").first()
        random_individual = get_patient_zero(
            self.population, PatientZeroStrategies.MOST_CONNECTIONS
        )
        random_individual.pathogen = self.pathogen
        random_individual.save()

    def test__run_simulation(self):
        """
        Mainly for me to try out shit
            - Should extract querying active infections to manager
            - Should filter away dead connections as well (prune)
            - Runtime should only be large for an interation if incubating + sick is large enough
            - Should fitler away recovered
        """
        from simulations.simulate import run_simulation, population_tick
        from tabulate import tabulate
        import time

        stats = []
        for _ in range(self.days):
            start = time.time()
            print(f"Running simulation for day: {_}")
            stat = population_tick(self.population)
            end = time.time()
            runtime = end - start
            stat["runtime"] = runtime
            stats.append(stat)

            if _ == 0:
                # No diff to calculate when initializing
                continue

            print(f"============Day {_}===============")
            print("--- Runtime: %s seconds ---" % runtime)
            print(
                tabulate(
                    [[k, v, v - stats[_ - 1][k]] for k, v in stat.items()],
                    headers=["Statistic", "Value", "Change"],
                    tablefmt="orgtbl",
                )
            )
            # create_text_bar_chart(stat)

            if stat["sick"] + stat["incubating"] == 0:
                print("No more sick or incubating")
                print("Exiting early")
                break

        runtime_list = [stat["runtime"] for stat in stats]
        print(f"Average runtime: {sum(runtime_list) / len(runtime_list)}")
        print(f"All runs: {runtime_list}")
        print(f"Total runtime: {sum(runtime_list)}")

        print("============ Main test loop done =================")
