import json
import numpy as np
from django.http import HttpResponse
from django.shortcuts import render, get_object_or_404, redirect
from simulations.models import Experiment, Generation, Connection, Individual, Pathogen
from django.db.models import Avg, FloatField
from django.db.models.functions import Coalesce
import matplotlib.pyplot as plt
import io
import urllib, base64
from django.http import FileResponse
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.figure import Figure

def index(request):
    return redirect("experiment_list")

def experiment_list(request):
    experiments = Experiment.objects.all().order_by("-started_at")
    return render(request, "experiment_list.html", {"experiments": experiments})

def experiment_detail_network_graph(request, pk):
    experiment = get_object_or_404(Experiment, id=pk)
    individuals = experiment.base_individual_network_topology.all()

    # Create a list of nodes from the individuals
    nodes = [{"id": str(ind.id)} for ind in individuals]

    # Get all connections between individuals associated with this experiment
    connections = Connection.objects.filter(individual__in=individuals)

    # Create a list of links from the connections
    links = [
        {"source": str(conn.individual.id), "target": str(conn.connected_individual.id), "value": conn.weight}
        for conn in connections
    ]
    return render(request, "experiment_detail_network_graph.html", {"nodes": json.dumps(nodes), "links": json.dumps(links)})


def experiment_delete(request, pk):
    if not request.method == "POST":
        return HttpResponse("Method not allowed", status=405)

    experiment = get_object_or_404(Experiment, id=pk)
    experiment.delete()
    return redirect("experiment_list")

def experiment_details(request, pk):
    experiment = get_object_or_404(Experiment, id=pk)

    experiment_generations = Generation.objects.filter(experiment=experiment)

    # Get all individuals associated with this experiment


    generation_batches = []
    graph_data = []
    for generation_index in range(1, experiment.number_of_generations + 1):
        current_generation = experiment_generations.filter(generation_number=generation_index)
        average_score = current_generation.aggregate(
            average_score=Coalesce(Avg("pathogen_performance_score"), 0, output_field=FloatField())
        )["average_score"]

        try:
            best_score = current_generation.order_by("-pathogen_performance_score").first().pathogen_performance_score
            worst_score = current_generation.order_by("pathogen_performance_score").first().pathogen_performance_score
        except AttributeError:
            best_score = 0
            worst_score = 0

        generation_batches.append(
            (
                generation_index,
                current_generation.order_by("pathogen_performance_score"),
                average_score,
            )
        )
        graph_data.append({
            'generation': generation_index,
            'average_score': average_score,
            'best_score': best_score,
            'worst_score': worst_score
        })
    ctx = {
        "experiment": experiment,
        "generation_batches": generation_batches,
        "graph_data": graph_data,  # Include the graph data in the context
    }

    return render(request, "experiment_details.html", ctx)

def pathogen_skill_point_distribution_graph(request, pk):
    pathogen = get_object_or_404(Pathogen, id=pk)

    # Get the point data
    infection_rate_points = pathogen.infection_rate_points
    death_points = pathogen.death_rate_points
    incubation_time_points = pathogen.incubation_time_points
    infection_time_points = pathogen.infection_time_points

    # Calculate averages
    avg_infection_rate = np.mean(infection_rate_points)
    avg_death_points = np.mean(death_points)
    avg_incubation_time = np.mean(incubation_time_points)
    avg_infection_time = np.mean(infection_time_points)

    fig = Figure(figsize=(10, 6))
    ax = fig.subplots()

    categories = ['Infection Rate', 'Death Rate', 'Incubation Time', 'Infection Time']
    values = [avg_infection_rate, avg_death_points, avg_incubation_time, avg_infection_time]

    # Create bar plot
    bars = ax.bar(categories, values)

    # naming the x axis
    ax.set_xlabel('Categories')
    # naming the y axis
    ax.set_ylabel('Average Value')

    # giving a title to my graph
    ax.set_title('Pathogen Skill Point Distribution')

    # Adding the value of each bar to the chart
    for bar in bars:
        y_val = bar.get_height()
        ax.text(bar.get_x() + bar.get_width() / 2.0, y_val, round(y_val, 2), va='bottom')  # va: vertical alignment

    # serve image
    canvas = FigureCanvas(fig)
    buf = io.BytesIO()
    canvas.print_png(buf)
    buf.seek(0)  # rewind the buffer to the start of the file
    return FileResponse(buf, as_attachment=False, filename='image.png', content_type='image/png')

